/* OthelloAI.java */

import java.awt.Point;

/** Othello computer player. */
public class OthelloAI
	implements OthelloPlayer, Runnable
{
	/** The player's color. */
	private int color;
	/** The AI level of calculation. */
	private int level;
	/** Othello interface. */
	private Othello o;
	/** Othello gameboard. */
	private OthelloBoard board;
	/** The thread that calculates the player's next move. */
	private Thread brain;
	
	/** Creates a computer player. */
	public OthelloAI(int color, int level, Othello o)
	{
		this.color=color;
		this.level=level;
		this.o=o;
	}
	
	/** Returns the player's color. */
	public int getColor()
	{
		return color;
	}
	
	/** Starts player's move.
	    This method will start a thinking thread,
	    that will eventually notify the Othello object of its move,
	    through the tryAndSet()-method. */
	public void startMove(OthelloBoard ob)
	{
		board=ob;
		brain=new Thread(this);
		brain.start();
	}
	
	/** Cancels player's move.
	    This method will not actually stop activity,
	    but it will prevent any gameboard values from being changed. */
	public void stopMove()
	{
		brain=null;
	}
	
	/** Runs thread.
	    This causes the */
	public void run()
	{
		Point [] p=board.canPut(color);
		if(p.length<1)
			return; // This should have been prevented, currently causes game to halt
		if(p.length>1)
		{
			// Search for the best places to place the chip
			int best=0, index[]=new int[1];
			for(int n=0;n<p.length;n++)
			{
				OthelloBoard ob=board.copy();
				ob.put(p[n].x, p[n].y, color);
				int score=getScore(ob, false, level);
				if(score>best)
				{
					// A new record
					best=score;
					index=new int[1];
					index[0]=n;
				}
				else if(score==best)
				{
					// This one was as good as a previous one,
					// and will be added to the queue.
					int [] temp=new int[index.length+1]; // realloc
					for(int i=0;i<index.length;i++)
						temp[i]=index[i];
					temp[index.length]=n;
					index=temp;
				}
			}
			if(brain!=null)
			{
				// Pick one of the best
				int i=index[(int)(Math.random()*(double)index.length)];
				o.tryToSet(p[i].x, p[i].y);
			}
		}
		else
			o.tryToSet(p[0].x, p[0].y);
	}
	
	/** Returns a gameboard score.
	    This score is based the number of chips of my color on the board,
	    minus the number of chips of the opponents color,
	    but extra points are given to border and corner squares. */
	private int getScore(OthelloBoard brd)
	{
		int score=0;
		for(int i=0;i<brd.getWidth();i++)
			for(int j=0;j<brd.getHeight();j++)
				if(brd.get(i, j)==color)
				{
					if(brd.isCorner(i, j))
						score+=5;
					else if(brd.isBorder(i, j))
						score+=2;
					else
						score+=1;
				}
				else if(brd.get(i, j)==brd.getOtherColor(color))
				{
					if(brd.isCorner(i, j))
						score-=5;
					else if(brd.isBorder(i, j))
						score-=2;
					else
						score-=1;
				}
		return score;
		//return brd.getNumber(color);
	}
	
	/** Returns a gameboard score.
	    This method recursivly seeks through all possible moves,
	    returning the "best case scenario" if it was my move,
	    and "worst case scenario" it it was the opponent's move. */
	private int getScore(OthelloBoard brd, boolean me, int lvl)
	{
		if(lvl==0)
			return getScore(brd);
		int c=(me?color:brd.getOtherColor(color));
		int best=(me?-1000:1000);
		Point [] p=brd.canPut(c);
		if(p.length<1)
		{
			OthelloBoard ob=brd.copy();
			best=getScore(ob, !me, lvl-1);
		}
		else
		{
			for(int n=0;n<p.length;n++)
			{
				OthelloBoard ob=brd.copy();
				ob.put(p[n].x, p[n].y, c);
				int score=getScore(ob, !me, lvl-1);
				if(me)
				{
					if(score>best)
						best=score;
				}
				else
				{
					if(score<best)
						best=score;
				}
			}
		}
		return best;
	}
}
