/* OthelloBoard */

import java.awt.Point;

/** Othello game board. */
public class OthelloBoard
{
	/** An empty square. */
	public static final int EMPTY=0;
	/** A black chip. */
	public static final int BLACK=1;
	/** A white chip. */
	public static final int WHITE=2;
	
	/** Gameboard width. */
	private final int width=8;
	/** Gameboard height. */
	private final int height=8;
	
	/** Gameboard. */
	private int data[]={
		EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, BLACK, WHITE, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, WHITE, BLACK, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY,
		EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY, EMPTY};
	
	/** Creates a new gameboard. */
	public OthelloBoard()
	{
	}
	
	/** Returns the contents of a square. */
	public int get(int u, int v)
	{
		if(u<0 || u>=width || v<0 || v>=height)
			return EMPTY;
		return data[u+v*width];
	}
	
	/** Returns the width of the gameboard. */
	public int getWidth()
	{
		return width;
	}
	
	/** Returns the height of the gameboard. */
	public int getHeight()
	{
		return height;
	}
	
	/** Places a chip on the gameboard.
	    Returns false if the chip cannot be placed there. */
	public boolean put(int u, int v, int color)
	{
		if(!canPut(u, v, color))
			return false;
		set(u, v, color);
		flip(u, v, color);
		return true;
	}
	
	/** Set the content of a square on the gameboard. */
	public void set(int u, int v, int color)
	{
		data[u+v*width]=color;
	}
	
	/** Flips chains of chips of the opposite color around (u, v). */
	private void flip(int u, int v, int color)
	{
		flip(getFlips(u, v, color));
	}
	
	/** Flips the squares adressed by the Point vector. */
	private void flip(Point [] p)
	{
		for(int i=0;i<p.length;i++)
			flip(p[i].x, p[i].y);
	}
	
	/** Flips the content of the square. */
	private void flip(int u, int v)
	{
		if(data[u+v*width]==WHITE)
			data[u+v*width]=BLACK;
		else if(data[u+v*width]==BLACK)
			data[u+v*width]=WHITE;
	}
	
	/** Returns the squares whose content may be flipped if a chip of the specified color is placed at (u,v).
	    If no such squares are detected, getFlips returns a vector of length 0. */
	public Point [] getFlips(int u, int v, int color)
	{
		final int [] dx={0, 1, 1, 1, 0, -1, -1, -1};
		final int [] dy={1, 1, 0, -1, -1, -1, 0, 1};
		PointVector pv=new PointVector();
		
		int ic=(color==BLACK?WHITE:BLACK);
		for(int dir=0;dir<8;dir++)
		{
			int x=u, y=v;
			int n=-1;
			do
			{
				x+=dx[dir];
				y+=dy[dir];
				n++;
			} while(get(x, y)==ic);
			if(get(x, y)==color)
			{
				x=u;
				y=v;
				for(int i=0;i<n;i++)
				{
					x+=dx[dir];
					y+=dy[dir];
					pv.add(x, y);
				}
			}
		}
		return pv.getVector();
	}
	
	/** Returns true if the chip can be placed in the square (u, v). */
	public boolean canPut(int u, int v, int color)
	{
		if(u<0 || u>=width || v<0 || v>=height)
			return false;
		if(data[u+v*width]!=EMPTY)
			return false;
		if(getFlips(u, v, color).length==0)
			return false;
		return true;
	}
	
	/** Returns a vector of squares where chips of the specified color may be placed. */
	public Point [] canPut(int color)
	{
		PointVector pv=new PointVector();
		for(int i=0;i<width;i++)
			for(int j=0;j<height;j++)
				if(canPut(i, j, color))
					pv.add(i, j);
		return pv.getVector();
	}
	
	/** Returns the number of squares holding chips of the specified color. */
	public int getNumber(int color)
	{
		int n=0;
		for(int i=0;i<width*height;i++)
			if(data[i]==color)
				n++;
		return n;
	}
	
	/** Returns the opposite color. */
	public int getOtherColor(int color)
	{
		return (color==BLACK?WHITE:(color==WHITE?BLACK:EMPTY));
	}
	
	/** Returns true if square (u, v) is a corner square. */
	public boolean isCorner(int x, int y)
	{
		return ((x==0 && y==0) || (x==width-1 && y==0) || (x==width-1 && y==height-1) || (x==0 && y==height-1));
	}
	
	/** Returns true if square (u, v) is a border square. */
	public boolean isBorder(int x, int y)
	{
		return ((x==0) || (y==0) || (x==width-1) || (y==height-1));
	}
	
	/** Returns a copy of the gameboard. */
	public OthelloBoard copy()
	{
		OthelloBoard ob=new OthelloBoard();
		for(int i=0;i<width;i++)
			for(int j=0;j<height;j++)
				ob.set(i, j, get(i,j));
		return ob;
	}
}

/** Point node. */
class PointNode
{
	/** A point. */
	Point p;
	/** The next node in a linked list. */
	PointNode next;
}

/** An abstract list of points. */
class PointVector
{
	/** Root node. */
	PointNode root;
	
	/** Create a list of Points. */
	public PointVector()
	{
		root=null;
	}
	
	/** Adds a point to the list. */
	public void add(int x, int y)
	{
		root=add(root, new Point(x, y));
	}
	
	/** Recursivly adds a point to the list, and returns the node that was added to. */
	private PointNode add(PointNode pn, Point p)
	{
		if(pn==null)
		{
			pn=new PointNode();
			pn.p=p;
		}
		else
			pn.next=add(pn.next, p);
		return pn;
	}
	
	/** Returns a vector of points.
	    If*/
	public Point [] getVector()
	{
		PointNode pn=root;
		int n=0;
		while(pn!=null)
		{
			pn=pn.next;
			n++;
		}
		
		Point [] p=new Point[n];
		
		pn=root;
		n=0;
		while(pn!=null)
		{
			p[n]=pn.p;
			pn=pn.next;
			n++;
		}
		return p;
	}
}
